/* RTOS C-level code including the scheduler function */
/* Carousel and cache variants */

#include "rtos.h"

#define MAX_NAME_LEN        8

typedef struct TCB_struct {
        void *      function;   /* offset 0 for asm */
        unsigned    running;    /* offset 4 for asm */
        unsigned    block_info_word; /* offset 8 for asm */
        unsigned    dynamic_release, dynamic_deadline, period;
        unsigned    offset;
        char        name[MAX_NAME_LEN+1];
} TCB;

static TCB thread_table[MAX_THREADS] AKD;
static unsigned num_threads AKD = 0;

/*asm(".section .kernelstuff.t,\"ax\",@progbits\n");*/

void register_task(const char * name, 
                void * function, unsigned block_info_word,
                unsigned offset, unsigned period) 
{
    TCB * p = &thread_table[num_threads + 1];

    if ((num_threads + 1) >= MAX_THREADS) {
        print("too many threads\n");
        exit(1);
    }
    if (!function) {
        print("null function registered!\n");
        exit(1);
    }
    num_threads++;
    p->function = function;
    p->block_info_word = block_info_word;
    p->offset = offset;
    p->period = period;
    memcpy(p->name, name, MAX_NAME_LEN);
    p->name[MAX_NAME_LEN] = '\0';
    notify_task_name(name, function);
}

void * scheduler(void) 
{
    /* called by ISR */
    unsigned i;
    unsigned now = current_time;
    unsigned preempt = 0xffffffff;
    TCB * best_task = &thread_table[0]; /* idle task, always runnable */

    /* Check deadlines */
    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];

        if (p->dynamic_deadline < now) {
            print("*** deadline overrun ***\n");
            for (i = num_threads; i > 0; i--) {
                p = &thread_table[i];
                if (p->dynamic_deadline < now) {
                    print(p->name);
                }
            }
            print("halted\n");
            exit(70);
        }
    }


    /* 1. What is the best_task to run now? */
    /* 2. At what time should we preempt? */

    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];
        
        if (p->running) {
            /* Do not reschedule; no new task */
            best_task = p;
            break;
        }

        if (p->dynamic_release <= now) {
            /* Runnable task */
            best_task = p;
            break;
        } else if (p->dynamic_release < preempt) {
            /* Higher-priority task, not yet runnable */
            preempt = p->dynamic_release;
        }
    }

    if (best_task->running) {
        /* No task switch this time; schedule preemption */
        current_time = preempt;
        return 0;
    } else {
        /* A new task starts; schedule preemption if any */
        current_time = preempt;

        best_task->dynamic_release += best_task->period;
        best_task->dynamic_deadline += best_task->period;
        best_task->running = 1;
        return (void *) best_task;
    }
}

extern void kernel_idle(void);

void crash(void) 
{
    print("crash called; idle was rescheduled?");
    exit(1);
}

void kernel_start(void) 
{
    unsigned now = current_time;
    unsigned i;

    /* initiate idle task */
    thread_table[0].function = crash;
    thread_table[0].running = 1;

    /* setup release times and deadlines */
    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];
        p->dynamic_release = p->offset + now;
        p->dynamic_deadline = p->period + p->offset + now;
        p->running = 0;
    }

    /* start timer, enable interrupts and go idle */
    kernel_idle();
}



